#!/usr/bin/env python3
"""Command-line interface to wavchunk"""
import argparse
import io
import os
import sys

import wavchunk

# -----------------------------------------------------------------------------


def main():
    """Main entry point"""
    args = get_args()

    # Dispatch to sub-command
    args.func(args)


# -----------------------------------------------------------------------------


def do_add(args):
    """Add chunk to WAV file"""
    in_file = sys.stdin.buffer
    out_file = sys.stdout.buffer
    data_file = None

    if args.data_file:
        # --data is a file path
        if args.data == "-":
            # Read data from stdin
            assert args.input, "Must specify --input if --data is stdin"
            data_file = sys.stdin.buffer

            if os.isatty(sys.stdin.fileno()):
                print("Reading data from stdin...", file=sys.stderr)
        else:
            # Read data from a file
            data_file = open(args.data, "rb")
    else:
        # Read data from argument
        data_file = io.BytesIO(args.data.encode())

    # Load chunk data
    chunk_data = data_file.read()

    if args.input:
        # Read WAV from file
        in_file = open(args.input, "rb")
    else:
        # Read WAV from stdin
        if os.isatty(sys.stdin.fileno()):
            print("Reading WAV from stdin...", file=sys.stderr)

    wavchunk.add_chunk(in_file, chunk_data, chunk_name=args.name, out_file=out_file)


# -----------------------------------------------------------------------------


def do_get(args):
    """Get chunk from WAV file"""
    in_file = sys.stdin.buffer
    out_file = sys.stdout.buffer
    data_file = None
    keep_chunk = True

    if args.delete:
        keep_chunk = False

    if args.data == "-":
        # Write data to stdout
        data_file = sys.stdout.buffer

        if not args.output:
            # No WAV output
            out_file = None
    elif args.data:
        # Write data to file
        data_file = open(args.data, "wb")

    if args.input:
        # Read WAV from stdin
        in_file = open(args.input, "rb")
    else:
        # Read WAV from a file
        if os.isatty(sys.stdin.fileno()):
            print("Reading WAV from stdin...", file=sys.stderr)

    if args.output:
        # Write WAV to stdout
        out_file = open(args.output, "wb")

    chunk_data = wavchunk.get_chunk(
        in_file, chunk_name=args.name, out_file=out_file, keep_chunk=keep_chunk
    )

    if chunk_data and data_file:
        data_file.write(chunk_data)


# -----------------------------------------------------------------------------


def get_args() -> argparse.Namespace:
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(prog="wavchunk")

    # Create subparsers for each sub-command
    sub_parsers = parser.add_subparsers()
    sub_parsers.required = True
    sub_parsers.dest = "command"

    # ---
    # add
    # ---
    add_parser = sub_parsers.add_parser("add", help="Add chunk to WAV file")
    add_parser.add_argument(
        "-d", "--data", required=True, help="Chunk data to add ('-' for stdin)"
    )
    add_parser.add_argument(
        "-i", "--input", help="Path to input WAV file (default: stdin)"
    )
    add_parser.add_argument(
        "--data-file", action="store_true", help="Argument to --data is a file"
    )
    add_parser.add_argument(
        "--name", default="INFO", help="Name of chunk to add (default: INFO)"
    )
    add_parser.set_defaults(func=do_add)

    # ---
    # get
    # ---
    get_parser = sub_parsers.add_parser("get", help="Get chunk data from WAV file")
    get_parser.add_argument(
        "-i", "--input", help="Path to input WAV file (default: stdin)"
    )
    get_parser.add_argument(
        "-d", "--data", help="Path to write chunk data ('-' for stdout)"
    )
    get_parser.add_argument(
        "-o", "--output", help="Path to write WAV data (default: stdout)"
    )
    get_parser.add_argument(
        "--delete", action="store_true", help="Remove chunk from WAV data"
    )
    get_parser.add_argument(
        "--name", default="INFO", help="Name of chunk to get (default: INFO)"
    )
    get_parser.set_defaults(func=do_get)

    return parser.parse_args()


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Tests for wavchunk"""
import io
import os
import unittest
import wave

import wavchunk


class WavChunkTestCase(unittest.TestCase):
    """Test cases for wavchunk"""

    def test_add_remove_chunk(self):
        """Test adding and removing INFO chunk"""
        # Generate random WAV file with 100 samples
        audio_data = os.urandom(2 * 100)

        with io.BytesIO() as wav_buffer:
            wav_file: wave.Wave_write = wave.open(wav_buffer, mode="wb")
            with wav_file:
                wav_file.setframerate(16000)
                wav_file.setsampwidth(2)
                wav_file.setnchannels(1)
                wav_file.writeframes(audio_data)

            wav_bytes = wav_buffer.getvalue()

        # Random data for INFO
        chunk_data = os.urandom(50)

        with io.BytesIO(wav_bytes) as in_file:
            with io.BytesIO() as out_file:
                wavchunk.add_chunk(in_file, chunk_data, out_file=out_file)
                out_bytes = out_file.getvalue()

        # Verify chunk data
        with io.BytesIO(out_bytes) as chunked_file:
            actual_data = wavchunk.get_chunk(chunked_file)
            self.assertEqual(chunk_data, actual_data)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



"""Utility methods for interacting with RIFF WAVE file chunks"""
import io
import shutil
import struct
import typing
from enum import Enum

# -----------------------------------------------------------------------------


class Names(str, Enum):
    """Chunk names in RIFF WAVE"""

    RIFF = "RIFF"
    WAVE = "WAVE"
    FMT = "fmt "
    INFO = "INFO"


# -----------------------------------------------------------------------------


def read_name(file: typing.IO[bytes]) -> str:
    """Read a 4 byte name from a RIFF file"""
    return file.read(4).decode()


def write_name(file: typing.IO[bytes], name: str):
    """Write a 4 byte name from a RIFF file"""
    while len(name) < 4:
        name += " "

    if len(name) > 4:
        raise ValueError(f"'{name}' must be exactly 4 characters long")

    file.write(name.encode())


def read_size(file: typing.IO[bytes]) -> int:
    """Read a 4 byte size from a RIFF file"""
    return struct.unpack_from("<L", file.read(4))[0]


def write_size(file: typing.IO[bytes], size: int):
    """Write a 4 byte size to a RIFF file"""
    file.write(struct.pack("<L", size))


def read_after_format(
    file: typing.IO[bytes], out_file: typing.Optional[typing.IO[bytes]] = None
):
    """
    Read until file is right after data chunk.

    Optionally copy data to output file.
    """
    # Copy RIFF header
    riff = read_name(file)
    assert riff == Names.RIFF, f"Expected {Names.RIFF}, got {riff}"
    riff_size = read_size(file)

    if out_file:
        write_name(out_file, riff)
        write_size(out_file, riff_size)

    # Copy WAVE header
    wave = read_name(file)
    assert wave == Names.WAVE, f"Expected {Names.WAVE}, got {wave}"

    if out_file:
        write_name(out_file, wave)

    # Format chunk
    fmt = read_name(file)
    assert fmt == Names.FMT, f"Expected '{Names.FMT}', got {fmt}"
    fmt_size = read_size(file)

    # Skip format chunk
    fmt_bytes = file.read(fmt_size)

    if out_file:
        write_name(out_file, fmt)
        write_size(out_file, fmt_size)
        out_file.write(fmt_bytes)


def read_after_data(
    file: typing.IO[bytes], out_file: typing.Optional[typing.IO[bytes]] = None
):
    """
    Read until file is right after data chunk.

    Optionally copy data to an output file.
    """
    # Skip until after format chunk
    read_after_format(file, out_file=out_file)

    # Skip chunks until after data
    last_name = ""

    while last_name != "data":
        chunk_name = read_name(file)
        chunk_size = read_size(file)
        chunk_bytes = file.read(chunk_size)

        if out_file:
            write_name(out_file, chunk_name)
            write_size(out_file, chunk_size)
            out_file.write(chunk_bytes)

        last_name = chunk_name


def add_chunk(
    wav_file: typing.IO[bytes],
    chunk_data: bytes,
    chunk_name: str = Names.INFO,
    out_file: typing.Optional[typing.IO[bytes]] = None,
) -> typing.Optional[bytes]:
    """
    Add new chunk after data (INFO by default).

    Optionally copy data to an output file.
    """
    if out_file:
        return_value = False
    else:
        out_file = io.BytesIO()
        return_value = True

    # Skip until after data
    read_after_data(wav_file, out_file=out_file)

    # Add new chunk
    write_name(out_file, chunk_name)
    write_size(out_file, len(chunk_data))
    out_file.write(chunk_data)

    # Copy remainder of input file
    shutil.copyfileobj(wav_file, out_file)

    if return_value:
        assert isinstance(out_file, io.BytesIO)
        return out_file.getvalue()

    return None


def get_chunk(
    wav_file: typing.IO[bytes],
    out_file: typing.Optional[typing.IO[bytes]] = None,
    chunk_name: str = Names.INFO,
    keep_chunk: bool = True,
) -> typing.Optional[bytes]:
    """
    Find a chunk and return data from it (default: INFO).


    Optionally copy data to an output file.
    Set keep_chunk = False to delete chunk.
    """
    return_data = None

    # Skip until after format chunk
    read_after_format(wav_file, out_file=out_file)

    # Read chunks until INFO is found
    current_name = read_name(wav_file)
    while current_name:
        chunk_size = read_size(wav_file)
        chunk_bytes = wav_file.read(chunk_size)

        if current_name == chunk_name:
            return_data = chunk_bytes
            if out_file and keep_chunk:
                # Copy chunk to output
                write_name(out_file, current_name)
                write_size(out_file, chunk_size)
                out_file.write(chunk_bytes)
                break
        elif out_file:
            # Copy chunk to output
            write_name(out_file, current_name)
            write_size(out_file, chunk_size)
            out_file.write(chunk_bytes)

        # Next chunk
        current_name = read_name(wav_file)

    if out_file:
        # Copy remainder of input file
        shutil.copyfileobj(wav_file, out_file)

    return return_data
